from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column
from sqlalchemy.types import Integer

maker = sessionmaker(autoflush=True, autocommit=False)
DBSession = scoped_session(maker)
DeclarativeBase = declarative_base()
metadata = DeclarativeBase.metadata 

class User(DeclarativeBase):
    __tablename__ = 'users'

    uid = Column(Integer, autoincrement=True, primary_key=True)

class FakeObject(DeclarativeBase):
    __tablename__ = 'object'

    uid = Column(Integer, autoincrement=True, primary_key=True)


from tw.api import Widget, JSLink, CSSLink
from tw.jquery import jquery_js
import tg

from tgext.tagging.model.sautils import get_primary_key

class TagList(Widget):
    template = "genshi:tgext.tagging.templates.taglist"
    javascript = [jquery_js, JSLink(modname='tgext.tagging', filename='static/tagging.js')]
    css = [CSSLink(modname='tgext.tagging', filename='static/tagging.css')]

    params = {'editmode': 'Enable Tags removal and adding',
              'tagging_url': 'Url of the tagging controller'}

    editmode = True
    tagging_url = '/tagging'

    def update_params(self, d):
        from tgext.tagging.model.tagging import Tagging
        
        super(TagList, self).update_params(d)
        d['tg_url'] = tg.url
        d['Tagging'] = Tagging

        if not isinstance(d['value'], dict):
            d['value'] = {'target_id': getattr(d['value'], get_primary_key(d['value'].__class__)),
                           'tag_list': Tagging.tag_cloud_for_object(d['value'])}
    
    



# See http://peak.telecommunity.com/DevCenter/setuptools#namespace-packages
try:
    __import__('pkg_resources').declare_namespace(__name__)
except ImportError:
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)


from nose.tools import raises
from sqlalchemy import Table, ForeignKey, Column, and_, create_engine
from sqlalchemy.types import Unicode, Integer
from model_meta import DeclarativeBase

from tgext.tagging.model.sautils import get_primary_key

class TestModel(DeclarativeBase):
    __tablename__ = 'test_model'

    pfield = Column(Integer, autoincrement=True, primary_key=True)

class NonIntegerPrimaryKey(DeclarativeBase):
    __tablename__ = 'test_model_with_wrong_key'

    name = Column(Unicode(127), primary_key=True)
    
class TestSAUtils(object):
    def test_primary_key(self):
        assert get_primary_key(TestModel) == 'pfield'

    @raises(LookupError)
    def test_model_missing_key(self):
        print get_primary_key(NonIntegerPrimaryKey)


from tw.api import Widget, JSLink, CSSLink
from tw.jquery import jquery_js
import tg

class TagCloud(Widget):
    template = "genshi:tgext.tagging.templates.tagcloud"
    javascript = [jquery_js, JSLink(modname='tgext.tagging', filename='static/tagging.js')]
    css = [CSSLink(modname='tgext.tagging', filename='static/tagging.css')]

    params = {'tagging_url': 'Url of the tagging controller'}

    tagging_url = '/tagging'

    def update_params(self, d):
        super(TagCloud, self).update_params(d)
        d['tg_url'] = tg.url
        d['weight_min'] = 0
        d['weight_max'] = max((t[1] for t in d['value']))
        d['weight_range'] = d['weight_max'] - d['weight_min']

        
    
    



from tgext.tagging.model.tagging import Tagging, Tag
from model_meta import DBSession, User, FakeObject

class TestTagging(object):
    def teardown(self):
        DBSession.rollback()

    def setup(self):
        DBSession.add(User())
        self.fake_o = FakeObject()
        DBSession.add(self.fake_o)
        self.fake_o2 = FakeObject()
        DBSession.add(self.fake_o2)
        self.fake_o3 = FakeObject()
        DBSession.add(self.fake_o3)
        DBSession.flush()
    
    def test_add_tags(self):
        Tagging.add_tags(self.fake_o, 'TaG1,tAg2')

        tags = DBSession.query(Tag).all()
        assert len(tags) == 2
        assert tags[0].name == 'tag1'
        assert tags[1].name == 'tag2'

    def test_repeated_add_tags(self):
        Tagging.add_tags(self.fake_o, 'TaG1,tAg2,tag2,TAG2')

        tags = DBSession.query(Tag).all()
        assert len(tags) == 2
        assert tags[0].name == 'tag1'
        assert tags[1].name == 'tag2'

    def test_tag_cloud(self):
        Tagging.add_tags(self.fake_o, 'TaG1,tAg2,tag2,TAG2')
        tag_cloud = Tagging.tag_cloud_for_object(self.fake_o)
        assert tag_cloud.count() == 2

    def test_tag_cloud_weights(self):
        Tagging.add_tags(self.fake_o, 'TaG1,tAg2,tag2,TAG2')
        Tagging.add_tags(self.fake_o2, 'tag1')

        tag_cloud = Tagging.tag_cloud_for_set(FakeObject, [self.fake_o, self.fake_o2])
        for tag in tag_cloud:
            if tag[0] == 'tag1':
                assert tag[1] == 2
            elif tag[0] == 'tag2':
                assert tag[1] == 1

    def test_tag_cloud_for_type(self):
        Tagging.add_tags(self.fake_o, 'tag1,tag2,tag3')
        Tagging.add_tags(self.fake_o2, 'tag1')
        Tagging.add_tags(self.fake_o3, 'tag1,tag2')

        tag_cloud = Tagging.tag_cloud_for_set(FakeObject)
        assert tag_cloud.count() == 3
        for tag in tag_cloud:
            if tag[0] == 'tag1':
                assert tag[1] == 3
            elif tag[0] == 'tag2':
                assert tag[1] == 2
            elif tag[0] == 'tag3':
                assert tag[1] == 1

    def test_del_tags(self):
        Tagging.add_tags(self.fake_o, 'TaG1,tAg2,tag2,TAG2')
        Tagging.del_tags(self.fake_o, 'tag2')

        tag_cloud = Tagging.tag_cloud_for_object(self.fake_o).all()
        assert len(tag_cloud) == 1
        assert tag_cloud[0].name == 'tag1'

    def test_set_tags(self):
        Tagging.add_tags(self.fake_o, 'TaG1,tAg2,tag2,TAG2')
        tag_cloud = Tagging.tag_cloud_for_object(self.fake_o).all()
        assert len(tag_cloud) == 2

        Tagging.set_tags(self.fake_o, 'tag1,TAG1')
        tag_cloud = Tagging.tag_cloud_for_object(self.fake_o).all()
        assert len(tag_cloud) == 1
        


from tgext.tagging.model.tagging import Tag
from model_meta import DBSession

class TestTags(object):
    def teardown(self):
        DBSession.rollback()
    
    def test_case_insensitive_tag(self):
        DBSession.add(Tag(name='Hi'))
        tag = Tag.lookup('hi')

        assert tag
        assert tag.name == 'hi'

    def test_case_insensitive_tag2(self):
        DBSession.add(Tag(name='hi'))
        tag = Tag.lookup('Hi')

        assert tag
        assert tag.name == 'hi'

    def test_list_lookup(self):
        DBSession.add(Tag(name='tag1'))
        DBSession.add(Tag(name='tag2'))

        tags = Tag.lookup_list('tag1, tag2, tag3')
        assert len(tags['found']) == 2
        assert len(tags['missing']) == 1



# -*- coding: utf-8 -*-
from tg import request

from datetime import datetime
from sqlalchemy import *
from sqlalchemy import Table, ForeignKey, Column, and_
from sqlalchemy.types import Unicode, Integer, DateTime, Boolean
from sqlalchemy.orm import backref, relation, synonym, MapperExtension
from sqlalchemy.ext.declarative import synonym_for

from metadata import DeclarativeBase, metadata, DBSession
from app_auth import User
from sautils import get_primary_key

import re
clean_up_regexp = re.compile('[^a-zA-Z0-9]')

class Tag(DeclarativeBase):
    class Hooks(MapperExtension):
        def before_insert(self, mapper, connection, instance):
            instance.name = clean_up_regexp.sub('', instance.name.strip().lower())

        def before_update(self, mapper, connection, instance):
            instance.name = clean_up_regexp.sub('', instance.name.strip().lower())

    __tablename__ = 'tgext_tagging_tags'
    __mapper_args__ = {'extension': Hooks()}


    id = Column(Integer, autoincrement=True, primary_key=True)
    name = Column(Unicode(255), nullable=False, unique=True, index=True)
    created_at = Column(DateTime, nullable=False, default=datetime.now)
    last_use = Column(DateTime, nullable=False, default=datetime.now)

    @classmethod
    def lookup(cls, tag_name):
        tag = DBSession.query(cls).filter_by(name=tag_name.strip().lower()).first()
        if tag:
            return tag

    @classmethod
    def lookup_list(cls, tag_list_as_string):
        tag_names = tag_list_as_string.split(',')
        tags = {'found':[], 'missing':[]}

        for tag in tag_names:
            tag = clean_up_regexp.sub('', tag.strip().lower())
            found = Tag.lookup(tag)
            if found:
                tags['found'].append(found)
            else:
                tags['missing'].append(tag)

        return tags


class Tagging(DeclarativeBase):
    __tablename__ = 'tgext_tagging_taggings'

    id = Column(Integer, autoincrement=True, primary_key=True)
    created_at = Column(DateTime, nullable=False, default=datetime.now)
    taggable_id = Column(Integer, nullable=False, index=True)
    taggable_type = Column(String(255), nullable=False, index=True)

    user_id = Column(Integer, ForeignKey(getattr(User, get_primary_key(User))),
                              nullable=True)
    user = relation(User, primaryjoin=user_id == getattr(User, get_primary_key(User)), uselist=False,
                            backref=backref("perfomed_tagging", order_by=desc('created_at'),
                                                                cascade='all, delete-orphan'))

    tag_id = Column(Integer, ForeignKey('tgext_tagging_tags.id'), nullable=False)
    tag = relation('Tag', backref=backref('taggings', cascade="all, delete-orphan"), uselist=False)

    @classmethod
    def items_for_tags(cls, Type, tags):
        type_primary_key = get_primary_key(Type)
        tags = tags.split(',')
        
        return DBSession.query(Type).join((Tagging, Tagging.taggable_id == getattr(Type, type_primary_key)))\
                                    .join((Tag, Tag.id == Tagging.tag_id))\
                                    .filter(Tagging.taggable_type == Type.__name__)\
                                    .filter(Tag.name.in_(tags))\
                                    .group_by(Tagging.taggable_id)\
                                    .having(func.count(Tagging.taggable_id) == len(tags))

    @classmethod
    def tag_cloud_for_user(cls, user, Type=None):
        q = DBSession.query(Tag.name, func.count('*').label('tags_count')).join(Tagging)\
                     .filter(Tagging.user == user)

        if Type:
            type_primary_key = get_primary_key(Type)
            q = q.join((Type, getattr(Type, type_primary_key) == Tagging.taggable_id))\
                 .filter(Tagging.taggable_type == Type.__name__)

        q = q.group_by(Tag.id)
        return q.order_by(Tag.name)

    @classmethod
    def tag_cloud_for_object(cls, entry):
        Type = entry.__class__
        type_primary_key = get_primary_key(Type)
        oid = getattr(entry, type_primary_key)
        
        q = DBSession.query(Tag.name, func.count('*').label('tags_count')).join(Tagging)\
                     .join((Type, getattr(Type, type_primary_key) == Tagging.taggable_id))\
                     .filter(getattr(Type, type_primary_key) == oid)\
                     .filter(Tagging.taggable_type == Type.__name__)\
                     .group_by(Tag.id)
        return q.order_by(Tag.name)

    @classmethod
    def tag_cloud_for_set(cls, Type, set=None):
        type_primary_key = get_primary_key(Type)

        if set:
            filter_on_items = [Tagging.taggable_id == getattr(o, type_primary_key) for o in set]
            if not filter_on_items:
                return []

        q = DBSession.query(Tag.name, func.count('*').label('tags_count'))\
                     .join(Tagging).filter(Tagging.taggable_type == Type.__name__)

        if set:
            q = q.filter(or_(*filter_on_items))

        q = q.group_by(Tag.id)
        return q.order_by(Tag.name)

    @classmethod
    def add_tags(cls, entry, tags, user=None):
        Type = entry.__class__
        oid = getattr(entry, get_primary_key(Type))
        
        tags = Tag.lookup_list(tags)
        tag_cloud = [x[0] for x in cls.tag_cloud_for_object(entry)]
            
        for tag in tags['found']:
            if tag.name not in tag_cloud and tag.name != '':
                try:
                    tagging = Tagging(taggable_type=Type.__name__, taggable_id=oid,
                                      user=user, tag=tag)
                    DBSession.add(tagging)
                    tag.last_use = datetime.now()
                except IntegrityError:
                    pass

        for tag in tags['missing']:
            if tag:
                check_just_been_added = Tag.lookup_list(tag)
                if check_just_been_added['found']:
                    continue

                tag = Tag(name=tag.strip().lower())
                DBSession.add(tag)
                DBSession.flush()
                try:
                    tagging = Tagging(taggable_type=Type.__name__, taggable_id=oid,
                                      user=user, tag=tag)
                    DBSession.add(tagging)
                except IntegrityError:
                    pass

    @classmethod
    def del_tags(cls, entry, tags):
        Type = entry.__class__
        oid = getattr(entry, get_primary_key(Type))
        
        tags = Tag.lookup_list(tags)
        tag_cloud = (x[0] for x in cls.tag_cloud_for_object(entry))

        for tag in tags['found']:
            if tag.name in tag_cloud:
                tagging = DBSession.query(Tagging).filter(Tagging.tag_id == tag.id)\
                                                  .filter(Tagging.taggable_type == Type.__name__)\
                                                  .filter(Tagging.taggable_id == oid).first()
                if tagging:
                    DBSession.delete(tagging)
 
    @classmethod
    def set_tags(cls, entry, tags):
        Type = entry.__class__
        oid = getattr(entry, get_primary_key(Type))

        item_tag_cloud = (x[0].lower() for x in cls.tag_cloud_for_object(entry))
        tags_to_set = list(set(x.strip().lower() for x in tags.split(',')))
        tags_to_remove = filter(lambda tag : tag not in tags_to_set, item_tag_cloud)

        cls.del_tags(entry, ",".join(tags_to_remove))
        cls.add_tags(entry, ",".join(tags_to_set))
